@@ -1,26 +1,26 @@
-import sys
-import os
-import argparse
 import logging
-import requests
-import jsonpickle
-from pathlib import Path
-from tqdm import tqdm
-from zipfile import ZipFile
-from distutils.dir_util import mkpath
-import operator
+import os
+import sys
 import pkgutil
 import importlib.machinery
 import importlib.util
 import types
+import inspect
 import shutil
 from copy import deepcopy
 import traceback
-import inspect
-import re
+import operator
+from pathlib import Path
 from subprocess import Popen
 from subprocess import PIPE
 from subprocess import STDOUT
+import argparse
+import requests
+import jsonpickle
+from tqdm import tqdm
+from zipfile import ZipFile
+from distutils.dir_util import mkpath
+import re
 
 ######## START CONTENT ########
 def INVALID_NAME():
@@ -165,165 +165,6 @@ class Datum(SelfRegistering):
 		this.valid = False
 
 
-# A DataContainer allows Data to be stored and worked with.
-# This class is intended to be derived from and added to.
-# Each DataContainer is comprised of multiple Data (see Datum.py for more).
-# NOTE: DataContainers are, themselves Data. Thus, you can nest your child classes however you would like.
-class DataContainer(Datum):
-
-	def __init__(this, name=INVALID_NAME()):
-		super().__init__(name)
-
-		# The data *this contains.
-		this.data = []
-
-
-	# RETURNS: an empty, invalid Datum.
-	def InvalidDatum(this):
-		ret = Datum()
-		ret.Invalidate()
-		return ret
-
-
-	# Sort things! Requires by be a valid attribute of all Data.
-	def SortData(this, by):
-		this.data.sort(key=operator.attrgetter(by))
-
-
-	# Adds a Datum to *this
-	def AddDatum(this, datum):
-		this.data.append(datum)
-
-
-	# RETURNS: a Datum with datumAttribute equal to match, an invalid Datum if none found.
-	def GetDatumBy(this, datumAttribute, match):
-		for d in this.data:
-			try: # within for loop 'cause maybe there's an issue with only 1 Datum and the rest are fine.
-				if (str(getattr(d, datumAttribute)) == str(match)):
-					return d
-			except Exception as e:
-				logging.error(f"{this.name} - {e.message}")
-				continue
-		return this.InvalidDatum()
-
-
-	# RETURNS: a Datum of the given name, an invalid Datum if none found.
-	def GetDatum(this, name):
-		return this.GetDatumBy('name', name)
-
-
-	# Removes all Data in toRem from *this.
-	# RETURNS: the Data removed
-	def RemoveData(this, toRem):
-		# logging.debug(f"Removing {toRem}")
-		this.data = [d for d in this.data if d not in toRem]
-		return toRem
-
-
-	# Removes all Data which match toRem along the given attribute
-	def RemoveDataBy(this, datumAttribute, toRem):
-		toRem = [d for d in this.data if str(getattr(d, datumAttribute)) in list(map(str, toRem))]
-		return this.RemoveData(toRem)
-
-
-	# Removes all Data in *this except toKeep.
-	# RETURNS: the Data removed
-	def KeepOnlyData(this, toKeep):
-		toRem = [d for d in this.data if d not in toKeep]
-		return this.RemoveData(toRem)
-
-
-	# Removes all Data except those that match toKeep along the given attribute
-	# RETURNS: the Data removed
-	def KeepOnlyDataBy(this, datumAttribute, toKeep):
-		# logging.debug(f"Keeping only class with a {datumAttribute} of {toKeep}")
-		# toRem = []
-		# for d in this.class:
-		#	 shouldRem = False
-		#	 for k in toKeep:
-		#		 if (str(getattr(d, datumAttribute)) == str(k)):
-		#			 logging.debug(f"found {k} in {d.__dict__}")
-		#			 shouldRem = True
-		#			 break
-		#	 if (shouldRem):
-		#		 toRem.append(d)
-		#	 else:
-		#		 logging.debug(f"{k} not found in {d.__dict__}")
-		toRem = [d for d in this.data if str(getattr(d, datumAttribute)) not in list(map(str, toKeep))]
-		return this.RemoveData(toRem)
-
-
-	# Removes all Data with the name "INVALID NAME"
-	# RETURNS: the removed Data
-	def RemoveAllUnlabeledData(this):
-		toRem = []
-		for d in this.data:
-			if (d.name =="INVALID NAME"):
-				toRem.append(d)
-		return this.RemoveData(toRem)
-
-
-	# Removes all invalid Data
-	# RETURNS: the removed Data
-	def RemoveAllInvalidData(this):
-		toRem = []
-		for d in this.data:
-			if (not d.IsValid()):
-				toRem.append(d)
-		return this.RemoveData(toRem)
-
-
-	# Removes all Data that have an attribute value relative to target.
-	# The given relation can be things like operator.le (i.e. <=)
-	#   See https://docs.python.org/3/library/operator.html for more info.
-	# If ignoreNames is specified, any Data of those names will be ignored.
-	# RETURNS: the Data removed
-	def RemoveDataRelativeToTarget(this, datumAttribute, relation, target, ignoreNames = []):
-		try:
-			toRem = []
-			for d in this.data:
-				if (ignoreNames and d.name in ignoreNames):
-					continue
-				if (relation(getattr(d, datumAttribute), target)):
-					toRem.append(d)
-			return this.RemoveData(toRem)
-		except Exception as e:
-			logging.error(f"{this.name} - {e.message}")
-			return []
-
-
-	# Removes any Data that have the same datumAttribute as a previous Datum, keeping only the first.
-	# RETURNS: The Data removed
-	def RemoveDuplicateDataOf(this, datumAttribute):
-		toRem = [] # list of Data
-		alreadyProcessed = [] # list of strings, not whatever datumAttribute is.
-		for d1 in this.data:
-			skip = False
-			for dp in alreadyProcessed:
-				if (str(getattr(d1, datumAttribute)) == dp):
-					skip = True
-					break
-			if (skip):
-				continue
-			for d2 in this.data:
-				if (d1 is not d2 and str(getattr(d1, datumAttribute)) == str(getattr(d2, datumAttribute))):
-					logging.info(f"Removing duplicate Datum {d2} with unique id {getattr(d2, datumAttribute)}")
-					toRem.append(d2)
-					alreadyProcessed.append(str(getattr(d1, datumAttribute)))
-		return this.RemoveData(toRem)
-
-
-	# Adds all Data from otherDataContainer to *this.
-	# If there are duplicate Data identified by the attribute preventDuplicatesOf, they are removed.
-	# RETURNS: the Data removed, if any.
-	def ImportDataFrom(this, otherDataContainer, preventDuplicatesOf=None):
-		this.data.extend(otherDataContainer.data);
-		if (preventDuplicatesOf is not None):
-			return this.RemoveDuplicateDataOf(preventDuplicatesOf)
-		return []
-
-
-
 # util is a namespace for any miscellaneous utilities.
 # You cannot create a util.
 class util:
@@ -842,10 +683,15 @@ class Functor(Datum):
 		for method in this.methods.values():
 			logging.debug(f"Populating method {this.name}.{method.name}({', '.join([a for a in method.requiredKWArgs] + [a+'='+str(v) for a,v in method.optionalKWArgs.items()])})")
 			method.object = this
-			setattr(this, method.name, method.__call__.__get__(this, this.__class__)) #...
+
+			# Python < 3.11
+			# setattr(this, method.name, method.__call__.__get__(this, this.__class__))
+
+			setattr(this, method.name, method.__call__.__get__(method, method.__class__))
 
 
-	# Set this.precursor
+
+# Set this.precursor
 	# Also set this.executor because it's easy.
 	def PopulatePrecursor(this):
 		if (this.executor is None):
@@ -927,6 +773,7 @@ class Functor(Datum):
 
 
 	# Hook for whatever logic you'd like to run before the next Functor is called.
+	# FIXME: This isn't actually called.
 	def PrepareNext(this, next):
 		pass
 
@@ -938,7 +785,7 @@ class Functor(Datum):
 			return None
 
 		if (this.GetExecutor() is None):
-			logging.warning(f"{this.name} has no executor and cannot execute next ({this.next}).")
+			raise InvalidNext(f"{this.name} has no executor and cannot execute next ({this.next}).")
 
 		next = this.next.pop(0)
 		if (not this.ValidateNext(next)):
@@ -1083,6 +930,457 @@ class Functor(Datum):
 
 	######## END: Fetch Locations ########
 
+
+def METHOD_PENDING_POPULATION(obj, *args, **kwargs):
+	raise MethodPendingPopulation("METHOD PENDING POPULATION")
+
+# Use the @method() decorator to turn any class function into an eons Method Functor.
+# Methods are Functors which can be implicitly transferred between classes (see Functor.PopulateMethods)
+# Using Methods also gives us full control over the execution of your code; meaning, we can change how Python interprets what you wrote!
+# All Methods will be stored in the method member of your Functor. However, you shouldn't need to access that.
+#
+# If you would like to specify a custom implementation, set the 'impl' kwarg (e.g. @method(impl='MyMethodImplementation'))
+# Beside 'impl', all key-word arguments provided to the @method() decorator will be set as member variables of the created Method.
+# For example, to set whether or not the Method should be propagated, you can use @method(propagate=True).
+# This means, you can create a custom means of interpreting your code with your own feature set and still use this @method decorator.
+# Perhaps you'd like something along the lines of: @method(impl='MakeAwesome', awesomeness=10000).
+# NOTE: in order for impl to work, the implementation class must already be Registered (or this must be called from an appropriate @recoverable function).
+def method(impl='Method', **kwargs):
+
+	# Class decorator with __set_name__, as described here: https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class
+	class MethodDecorator:
+		def __init__(this, function):
+			this.function = function
+
+		# Apparently, this is called when the decorated function is constructed.
+		def __set_name__(this, cls, functionName):
+			logging.debug(f"Constructing new method for {this.function} in {cls}")
+
+			# Create and configure a new Method
+
+			method = SelfRegistering(impl)
+			method.Constructor(this.function, cls)
+			for key, value in kwargs.items():
+				setattr(method, key, value)
+
+			# Store the new method in the class
+			if (not hasattr(cls, 'classMethods') or not isinstance(cls.classMethods, dict)):
+				cls.classMethods = {}
+			cls.classMethods[functionName] = method
+
+			# Self-destruct by replacing the decorated function.
+			# We rely on Functor.PopulateMethods to re-establish the method as callable.
+			# It seems like this just outright removes the methods. There may be an issue with using __get__ this way.
+			# Regardless deleting the method is okay as long as we add it back before anyone notices.
+			setattr(cls, functionName, METHOD_PENDING_POPULATION.__get__(cls))
+
+	return MethodDecorator
+
+class Method(Functor):
+
+	def __init__(this, name=INVALID_NAME()):
+		super().__init__(name)
+
+		# Whether or not *this should be combined with other Methods of the same name.
+		this.inheritMethods = True
+
+		# Where should inherited methods be inserted?
+		# First here means "before *this".
+		# If False, inherited code will be run after *this.
+		this.inheritedMethodsFirst = True # otherwise ...Last
+
+		# Propagation allows for Functors called after that which defines *this to also call *this.
+		# This system allows for partial, implicit inheritance.
+		# By default, Methods will not be propagated. Use @propagate to enable propagation.
+		this.propagate = False
+
+		# We don't care about these checks right now.
+		# Plus, we can't exactly wrap 2 functions even if we wanted to Rollback.
+		this.functionSucceeded = True
+		this.rollbackSucceeded = True
+		this.enableRollback = False
+
+		# The source code of the function we're implementing.
+		this.source = ""
+
+		# The instance of the class to which *this belongs.
+		# i.e. the object that called *this, aka 'owner', 'caller', etc.
+		this.object = None
+
+		this.original = util.DotDict()
+		this.original.cls = None
+		this.original.function = None
+
+
+	# Make *this execute the code in this.source
+	def UpdateSource(this):
+		wrappedFunctionName = f'_eons_method_{this.name}'
+		completeSource = f'''\
+def {wrappedFunctionName}(this):
+{this.source}
+'''
+		if (this.executor and this.executor.verbosity > 3):
+			logging.debug(f"Source for {this.name} is:\n{completeSource}")
+		code = compile(completeSource, '', 'exec')
+		exec(code)
+		exec(f'this.Function = {wrappedFunctionName}.__get__(this, this.__class__)')
+
+	# Parse arguments and update the source code
+	# TODO: Implement full python parser to avoid bad string replacement (e.g. "def func(self):... self.selfImprovement" => "... this.object.this.object.Improvement").
+	def PopulateFrom(this, function):
+		this.source = ':'.join(inspect.getsource(function).split(':')[1:]) #Remove the first function definition
+
+		args = inspect.signature(function, follow_wrapped=False).parameters
+		thisSymbol = next(iter(args))
+		#del args[thisSymbol] # ??? 'mappingproxy' object does not support item deletion
+		this.source = this.source.replace(thisSymbol, 'this.object')
+
+		first = True
+		for arg in args.values(): #args.values[1:] also doesn't work.
+			if (first):
+				first = False
+				continue
+
+			replaceWith = arg.name
+
+			if (arg.kind == inspect.Parameter.VAR_POSITIONAL):
+				replaceWith = 'this.args'
+
+			elif (arg.kind == inspect.Parameter.VAR_KEYWORD):
+				replaceWith = 'this.kwargs'
+
+			else:
+				if (arg.default != inspect.Parameter.empty):
+					this.optionalKWArgs[arg.name] = arg.default
+				else:
+					this.requiredKWArgs.append(arg.name)
+				replaceWith = f'this.{arg.name}'
+
+				if (arg.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.POSITIONAL_ONLY]):
+					this.argMapping.append(arg.name)
+
+			this.source = this.source.replace(arg.name, replaceWith)
+
+
+	# When properly constructing a Method, rely only on the function *this should implement.
+	# The class and all other properties are irrelevant. However, they are provided and intended for debugging only.
+	def Constructor(this, function, cls):
+		this.name = function.__name__
+		this.original.cls = cls
+		this.original.function = function
+
+		this.PopulateFrom(function)
+		this.UpdateSource()
+
+
+	# Grab any known and necessary args from this.kwargs before any Fetch calls are made.
+	def PopulatePrecursor(this):
+		if (not this.object):
+			raise MissingArgumentError(f"Call {this.name} from a class instance: {this.original.cls.__name__}.{this.name}(...). Maybe Functor.PopulateMethods() hasn't been called yet?")
+
+		this.executor = this.object.executor
+
+		if ('precursor' in this.kwargs):
+			this.precursor = this.kwargs.pop('precursor')
+		else:
+			this.precursor = None
+
+
+	# Next is set by Functor.PopulateMethods.
+	# We  definitely don't want to Fetch 'next'.
+	def PopulateNext(this):
+		pass
+
+
+	# Method.next should be a list of other Methods, as opposed to the standard string; so, instead of Executor.Execute..., we can directly invoke whatever is next.
+	# We skip all validation here.
+	# We also don't pass any args that were given in the initial function call. Those can all be Fetched from 'precursor'.
+	def CallNext(this):
+		if (not this.next):
+			return None
+
+		for next in this.next:
+			next(precursor=this)
+
+
+# A DataContainer allows Data to be stored and worked with.
+# This class is intended to be derived from and added to.
+# Each DataContainer is comprised of multiple Data (see Datum.py for more).
+# NOTE: DataContainers are, themselves Data. Thus, you can nest your child classes however you would like.
+class DataContainer(Datum):
+
+	def __init__(this, name=INVALID_NAME()):
+		super().__init__(name)
+
+		# The data *this contains.
+		this.data = []
+
+
+	# RETURNS: an empty, invalid Datum.
+	def InvalidDatum(this):
+		ret = Datum()
+		ret.Invalidate()
+		return ret
+
+
+	# Sort things! Requires by be a valid attribute of all Data.
+	def SortData(this, by):
+		this.data.sort(key=operator.attrgetter(by))
+
+
+	# Adds a Datum to *this
+	def AddDatum(this, datum):
+		this.data.append(datum)
+
+
+	# RETURNS: a Datum with datumAttribute equal to match, an invalid Datum if none found.
+	def GetDatumBy(this, datumAttribute, match):
+		for d in this.data:
+			try: # within for loop 'cause maybe there's an issue with only 1 Datum and the rest are fine.
+				if (str(getattr(d, datumAttribute)) == str(match)):
+					return d
+			except Exception as e:
+				logging.error(f"{this.name} - {e.message}")
+				continue
+		return this.InvalidDatum()
+
+
+	# RETURNS: a Datum of the given name, an invalid Datum if none found.
+	def GetDatum(this, name):
+		return this.GetDatumBy('name', name)
+
+
+	# Removes all Data in toRem from *this.
+	# RETURNS: the Data removed
+	def RemoveData(this, toRem):
+		# logging.debug(f"Removing {toRem}")
+		this.data = [d for d in this.data if d not in toRem]
+		return toRem
+
+
+	# Removes all Data which match toRem along the given attribute
+	def RemoveDataBy(this, datumAttribute, toRem):
+		toRem = [d for d in this.data if str(getattr(d, datumAttribute)) in list(map(str, toRem))]
+		return this.RemoveData(toRem)
+
+
+	# Removes all Data in *this except toKeep.
+	# RETURNS: the Data removed
+	def KeepOnlyData(this, toKeep):
+		toRem = [d for d in this.data if d not in toKeep]
+		return this.RemoveData(toRem)
+
+
+	# Removes all Data except those that match toKeep along the given attribute
+	# RETURNS: the Data removed
+	def KeepOnlyDataBy(this, datumAttribute, toKeep):
+		# logging.debug(f"Keeping only class with a {datumAttribute} of {toKeep}")
+		# toRem = []
+		# for d in this.class:
+		#	 shouldRem = False
+		#	 for k in toKeep:
+		#		 if (str(getattr(d, datumAttribute)) == str(k)):
+		#			 logging.debug(f"found {k} in {d.__dict__}")
+		#			 shouldRem = True
+		#			 break
+		#	 if (shouldRem):
+		#		 toRem.append(d)
+		#	 else:
+		#		 logging.debug(f"{k} not found in {d.__dict__}")
+		toRem = [d for d in this.data if str(getattr(d, datumAttribute)) not in list(map(str, toKeep))]
+		return this.RemoveData(toRem)
+
+
+	# Removes all Data with the name "INVALID NAME"
+	# RETURNS: the removed Data
+	def RemoveAllUnlabeledData(this):
+		toRem = []
+		for d in this.data:
+			if (d.name =="INVALID NAME"):
+				toRem.append(d)
+		return this.RemoveData(toRem)
+
+
+	# Removes all invalid Data
+	# RETURNS: the removed Data
+	def RemoveAllInvalidData(this):
+		toRem = []
+		for d in this.data:
+			if (not d.IsValid()):
+				toRem.append(d)
+		return this.RemoveData(toRem)
+
+
+	# Removes all Data that have an attribute value relative to target.
+	# The given relation can be things like operator.le (i.e. <=)
+	#   See https://docs.python.org/3/library/operator.html for more info.
+	# If ignoreNames is specified, any Data of those names will be ignored.
+	# RETURNS: the Data removed
+	def RemoveDataRelativeToTarget(this, datumAttribute, relation, target, ignoreNames = []):
+		try:
+			toRem = []
+			for d in this.data:
+				if (ignoreNames and d.name in ignoreNames):
+					continue
+				if (relation(getattr(d, datumAttribute), target)):
+					toRem.append(d)
+			return this.RemoveData(toRem)
+		except Exception as e:
+			logging.error(f"{this.name} - {e.message}")
+			return []
+
+
+	# Removes any Data that have the same datumAttribute as a previous Datum, keeping only the first.
+	# RETURNS: The Data removed
+	def RemoveDuplicateDataOf(this, datumAttribute):
+		toRem = [] # list of Data
+		alreadyProcessed = [] # list of strings, not whatever datumAttribute is.
+		for d1 in this.data:
+			skip = False
+			for dp in alreadyProcessed:
+				if (str(getattr(d1, datumAttribute)) == dp):
+					skip = True
+					break
+			if (skip):
+				continue
+			for d2 in this.data:
+				if (d1 is not d2 and str(getattr(d1, datumAttribute)) == str(getattr(d2, datumAttribute))):
+					logging.info(f"Removing duplicate Datum {d2} with unique id {getattr(d2, datumAttribute)}")
+					toRem.append(d2)
+					alreadyProcessed.append(str(getattr(d1, datumAttribute)))
+		return this.RemoveData(toRem)
+
+
+	# Adds all Data from otherDataContainer to *this.
+	# If there are duplicate Data identified by the attribute preventDuplicatesOf, they are removed.
+	# RETURNS: the Data removed, if any.
+	def ImportDataFrom(this, otherDataContainer, preventDuplicatesOf=None):
+		this.data.extend(otherDataContainer.data);
+		if (preventDuplicatesOf is not None):
+			return this.RemoveDuplicateDataOf(preventDuplicatesOf)
+		return []
+
+
+
+# The standard Functor extends Functor to add a set of standard members and methods.
+# This is similar to the standard library in C and C++
+# You must inherit from *this if you would like to use the functionality *this provides. The methods defined will not be propagated.
+class StandardFunctor(Functor):
+	def __init__(this, name="Standard Functor"):
+		super().__init__(name)
+
+	# Override this and do whatever!
+	# This is purposefully vague.
+	def Function(this):
+		pass
+
+
+	# Undo any changes made by UserFunction.
+	# Please override this too!
+	def Rollback(this):
+		pass
+
+
+	# Override this to check results of operation and report on status.
+	# Override this to perform whatever success checks are necessary.
+	def DidFunctionSucceed(this):
+		return this.functionSucceeded
+
+
+	# RETURN whether or not the Rollback was successful.
+	# Override this to perform whatever success checks are necessary.
+	def DidRollbackSucceed(this):
+		return this.rollbackSucceeded
+
+
+	######## START: UTILITIES ########
+
+	# RETURNS: an opened file object for writing.
+	# Creates the path if it does not exist.
+	@method()
+	def CreateFile(this, file, mode="w+"):
+		Path(os.path.dirname(os.path.abspath(file))).mkdir(parents=True, exist_ok=True)
+		return open(file, mode)
+
+	# Copy a file or folder from source to destination.
+	# This really shouldn't be so hard...
+	# root allows us to interpret '/' as something other than the top of the filesystem.
+	@method()
+	def Copy(this, source, destination, root='/'):
+		if (source.startswith('/')):
+			source = str(Path(root).joinpath(source[1:]).resolve())
+		else:
+			source = str(Path(source).resolve())
+		
+		destination = str(Path(destination).resolve())
+		
+		Path(destination).parent.mkdir(parents=True, exist_ok=True)
+
+		if (os.path.isfile(source)):
+			logging.debug(f"Copying file {source} to {destination}")
+			try:
+				shutil.copy(source, destination)
+			except shutil.Error as exc:
+				errors = exc.args[0]
+				for error in errors:
+					src, dst, msg = error
+					logging.debug(f"{msg}")
+		elif (os.path.isdir(source)):
+			logging.debug(f"Copying directory {source} to {destination}")
+			try:
+				shutil.copytree(source, destination)
+			except shutil.Error as exc:
+				errors = exc.args[0]
+				for error in errors:
+					src, dst, msg = error
+					logging.debug(f"{msg}")
+		else:
+			logging.error(f"Could not find source to copy: {source}")
+
+	# Delete a file or folder
+	@method()
+	def Delete(this, target):
+		if (not os.path.exists(target)):
+			logging.debug(f"Unable to delete nonexistent target: {target}")
+			return
+		if (os.path.isfile(target)):
+			logging.debug(f"Deleting file {target}")
+			os.remove(target)
+		elif (os.path.isdir(target)):
+			logging.debug(f"Deleting directory {target}")
+			try:
+				shutil.rmtree(target)
+			except shutil.Error as exc:
+				errors = exc.args[0]
+				for error in errors:
+					src, dst, msg = error
+					logging.debug(f"{msg}")
+
+	# Run whatever.
+	# DANGEROUS!!!!!
+	# RETURN: Return value and, optionally, the output as a list of lines.
+	@method()
+	def RunCommand(this, command, saveout=False, raiseExceptions=True):
+		logging.debug(f"================ Running command: {command} ================")
+		p = Popen(command, stdout=PIPE, stderr=STDOUT, shell=True)
+		output = []
+		while p.poll() is None:
+			line = p.stdout.readline().decode('utf8')[:-1]
+			if (saveout):
+				output.append(line)
+			if (line):
+				logging.debug(f"| {line}")  # [:-1] to strip excessive new lines.
+
+		if (p.returncode is not None and p.returncode):
+			raise CommandUnsuccessful(f"Command returned {p.returncode}")
+		
+		logging.debug(f"================ Completed command: {command} ================")
+		if (saveout):
+			return p.returncode, output
+		
+		return p.returncode
+	######## END: UTILITIES ########
+
 #from .Executor import Executor # don't import this, it'll be circular!
 
 # @recoverable
@@ -1643,297 +1941,6 @@ class Executor(DataContainer, Functor):
 	######## END: Fetch Locations ########
 
 
-def METHOD_PENDING_POPULATION(obj, *args, **kwargs):
-	raise MethodPendingPopulation("METHOD PENDING POPULATION")
-
-# Use the @method() decorator to turn any class function into an eons Method Functor.
-# Methods are Functors which can be implicitly transferred between classes (see Functor.PopulateMethods)
-# Using Methods also gives us full control over the execution of your code; meaning, we can change how Python interprets what you wrote!
-# All Methods will be stored in the method member of your Functor. However, you shouldn't need to access that.
-#
-# If you would like to specify a custom implementation, set the 'impl' kwarg (e.g. @method(impl='MyMethodImplementation'))
-# Beside 'impl', all key-word arguments provided to the @method() decorator will be set as member variables of the created Method.
-# For example, to set whether or not the Method should be propagated, you can use @method(propagate=True).
-# This means, you can create a custom means of interpreting your code with your own feature set and still use this @method decorator.
-# Perhaps you'd like something along the lines of: @method(impl='MakeAwesome', awesomeness=10000).
-# NOTE: in order for impl to work, the implementation class must already be Registered (or this must be called from an appropriate @recoverable function).
-def method(impl='Method', **kwargs):
-
-	# Class decorator with __set_name__, as described here: https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class
-	class MethodDecorator:
-		def __init__(this, function):
-			this.function = function
-
-		# Apparently, this is called when the decorated function is constructed.
-		def __set_name__(this, cls, functionName):
-			logging.debug(f"Constructing new method for {this.function} in {cls}")
-
-			# Create and configure a new Method
-
-			method = SelfRegistering(impl)
-			method.Constructor(this.function, cls)
-			for key, value in kwargs.items():
-				setattr(method, key, value)
-
-			# Store the new method in the class
-			if (not hasattr(cls, 'classMethods') or not isinstance(cls.classMethods, dict)):
-				cls.classMethods = {}
-			cls.classMethods[functionName] = method
-
-			# Self-destruct by replacing the decorated function.
-			# We rely on Functor.PopulateMethods to re-establish the method as callable.
-			# It seems like this just outright removes the methods. There may be an issue with using __get__ this way.
-			# Regardless deleting the method is okay as long as we add it back before anyone notices.
-			setattr(cls, functionName, METHOD_PENDING_POPULATION.__get__(cls))
-
-	return MethodDecorator
-
-class Method(Functor):
-
-	def __init__(this, name=INVALID_NAME()):
-		super().__init__(name)
-
-		# Whether or not *this should be combined with other Methods of the same name.
-		this.inheritMethods = True
-
-		# Where should inherited methods be inserted?
-		# First here means "before *this".
-		# If False, inherited code will be run after *this.
-		this.inheritedMethodsFirst = True # otherwise ...Last
-
-		# Propagation allows for Functors called after that which defines *this to also call *this.
-		# This system allows for partial, implicit inheritance.
-		# By default, Methods will not be propagated. Use @propagate to enable propagation.
-		this.propagate = False
-
-		# We don't care about these checks right now.
-		# Plus, we can't exactly wrap 2 functions even if we wanted to Rollback.
-		this.functionSucceeded = True
-		this.rollbackSucceeded = True
-		this.enableRollback = False
-
-		# The source code of the function we're implementing.
-		this.source = ""
-
-		# The instance of the class to which *this belongs.
-		# i.e. the object that called *this, aka 'owner', 'caller', etc.
-		this.object = None
-
-		this.original = util.DotDict()
-		this.original.cls = None
-		this.original.function = None
-
-
-	# Make *this execute the code in this.source
-	def UpdateSource(this):
-		wrappedFunctionName = f'_eons_method_{this.name}'
-		completeSource = f'''\
-def {wrappedFunctionName}(this):
-{this.source}
-'''
-		logging.debug(f"Source for {this.name} is:\n{completeSource}")
-		code = compile(completeSource, '', 'exec')
-		exec(code)
-		exec(f'this.Function = {wrappedFunctionName}.__get__(this, this.__class__)')
-
-	# Parse arguments and update the source code
-	# TODO: Implement full python parser to avoid bad string replacement (e.g. "def func(self):... self.selfImprovement" => "... this.object.this.object.Improvement").
-	def PopulateFrom(this, function):
-		this.source = ':'.join(inspect.getsource(function).split(':')[1:]) #Remove the first function definition
-
-		args = inspect.signature(function, follow_wrapped=False).parameters
-		thisSymbol = next(iter(args))
-		#del args[thisSymbol] # ??? 'mappingproxy' object does not support item deletion
-		this.source = this.source.replace(thisSymbol, 'this.object')
-
-		first = True
-		for arg in args.values(): #args.values[1:] also doesn't work.
-			if (first):
-				first = False
-				continue
-
-			replaceWith = arg.name
-
-			if (arg.kind == inspect.Parameter.VAR_POSITIONAL):
-				replaceWith = 'this.args'
-
-			elif (arg.kind == inspect.Parameter.VAR_KEYWORD):
-				replaceWith = 'this.kwargs'
-
-			else:
-				if (arg.default != inspect.Parameter.empty):
-					this.optionalKWArgs[arg.name] = arg.default
-				else:
-					this.requiredKWArgs.append(arg.name)
-				replaceWith = f'this.{arg.name}'
-
-				if (arg.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.POSITIONAL_ONLY]):
-					this.argMapping.append(arg.name)
-
-			this.source = this.source.replace(arg.name, replaceWith)
-
-
-	# When properly constructing a Method, rely only on the function *this should implement.
-	# The class and all other properties are irrelevant. However, they are provided and intended for debugging only.
-	def Constructor(this, function, cls):
-		this.name = function.__name__
-		this.original.cls = cls
-		this.original.function = function
-
-		this.PopulateFrom(function)
-		this.UpdateSource()
-
-
-	# Grab any known and necessary args from this.kwargs before any Fetch calls are made.
-	def PopulatePrecursor(this):
-		if (not this.object):
-			raise MissingArgumentError(f"Call {this.name} from a class instance: {this.original.cls.__name__}.{this.name}(...). Maybe Functor.PopulateMethods() hasn't been called yet?")
-
-		this.executor = this.object.executor
-
-		if ('precursor' in this.kwargs):
-			this.precursor = this.kwargs.pop('precursor')
-		else:
-			this.precursor = None
-
-
-	# Next is set by Functor.PopulateMethods.
-	# We  definitely don't want to Fetch 'next'.
-	def PopulateNext(this):
-		pass
-
-
-	# Method.next should be a list of other Methods, as opposed to the standard string; so, instead of Executor.Execute..., we can directly invoke whatever is next.
-	# We skip all validation here.
-	# We also don't pass any args that were given in the initial function call. Those can all be Fetched from 'precursor'.
-	def CallNext(this):
-		if (not this.next):
-			return None
-
-		for next in this.next:
-			next(precursor=this)
-
-
-# The standard Functor extends Functor to add a set of standard members and methods.
-# This is similar to the standard library in C and C++
-# You must inherit from *this if you would like to use the functionality *this provides. The methods defined will not be propagated.
-class StandardFunctor(Functor):
-	def __init__(this, name="Standard Functor"):
-		super().__init__(name)
-
-	# Override this and do whatever!
-	# This is purposefully vague.
-	def Function(this):
-		pass
-
-
-	# Undo any changes made by UserFunction.
-	# Please override this too!
-	def Rollback(this):
-		pass
-
-
-	# Override this to check results of operation and report on status.
-	# Override this to perform whatever success checks are necessary.
-	def DidFunctionSucceed(this):
-		return this.functionSucceeded
-
-
-	# RETURN whether or not the Rollback was successful.
-	# Override this to perform whatever success checks are necessary.
-	def DidRollbackSucceed(this):
-		return this.rollbackSucceeded
-
-
-	######## START: UTILITIES ########
-
-	# RETURNS: an opened file object for writing.
-	# Creates the path if it does not exist.
-	@method()
-	def CreateFile(this, file, mode="w+"):
-		Path(os.path.dirname(os.path.abspath(file))).mkdir(parents=True, exist_ok=True)
-		return open(file, mode)
-
-	# Copy a file or folder from source to destination.
-	# This really shouldn't be so hard...
-	# root allows us to interpret '/' as something other than the top of the filesystem.
-	@method()
-	def Copy(this, source, destination, root='/'):
-		if (source.startswith('/')):
-			source = str(Path(root).joinpath(source[1:]).resolve())
-		else:
-			source = str(Path(source).resolve())
-		
-		destination = str(Path(destination).resolve())
-		
-		Path(destination).parent.mkdir(parents=True, exist_ok=True)
-
-		if (os.path.isfile(source)):
-			logging.debug(f"Copying file {source} to {destination}")
-			try:
-				shutil.copy(source, destination)
-			except shutil.Error as exc:
-				errors = exc.args[0]
-				for error in errors:
-					src, dst, msg = error
-					logging.debug(f"{msg}")
-		elif (os.path.isdir(source)):
-			logging.debug(f"Copying directory {source} to {destination}")
-			try:
-				shutil.copytree(source, destination)
-			except shutil.Error as exc:
-				errors = exc.args[0]
-				for error in errors:
-					src, dst, msg = error
-					logging.debug(f"{msg}")
-		else:
-			logging.error(f"Could not find source to copy: {source}")
-
-	# Delete a file or folder
-	@method()
-	def Delete(this, target):
-		if (not os.path.exists(target)):
-			logging.debug(f"Unable to delete nonexistent target: {target}")
-			return
-		if (os.path.isfile(target)):
-			logging.debug(f"Deleting file {target}")
-			os.remove(target)
-		elif (os.path.isdir(target)):
-			logging.debug(f"Deleting directory {target}")
-			try:
-				shutil.rmtree(target)
-			except shutil.Error as exc:
-				errors = exc.args[0]
-				for error in errors:
-					src, dst, msg = error
-					logging.debug(f"{msg}")
-
-	# Run whatever.
-	# DANGEROUS!!!!!
-	# RETURN: Return value and, optionally, the output as a list of lines.
-	@method()
-	def RunCommand(this, command, saveout=False, raiseExceptions=True):
-		logging.debug(f"================ Running command: {command} ================")
-		p = Popen(command, stdout=PIPE, stderr=STDOUT, shell=True)
-		output = []
-		while p.poll() is None:
-			line = p.stdout.readline().decode('utf8')[:-1]
-			if (saveout):
-				output.append(line)
-			if (line):
-				logging.debug(f"| {line}")  # [:-1] to strip excessive new lines.
-
-		if (p.returncode is not None and p.returncode):
-			raise CommandUnsuccessful(f"Command returned {p.returncode}")
-		
-		logging.debug(f"================ Completed command: {command} ================")
-		if (saveout):
-			return p.returncode, output
-		
-		return p.returncode
-	######## END: UTILITIES ########
-
-
 # Use an ErrorStringParser for each "parsers" in order to avoid having to override the GetObjectFromError method and create a new class for every error you want to handle.
 # ErrorStringParsers enable ErrorResolutions to be created on a per-functionality, rather than per-error basis, reducing the total amount of duplicate code.
 # Each error has a different string. In order to get the object of the error, we have to know where the object starts and ends.
